#include "Vehicle.h"

#include <Aurora/Graphics/RenderManager.h>

#define DEFAULT_VEHICLE_SETUP
#define DEBUG_VEHICLE_WHEEL

Vehicle::Vehicle(void)
{
	mChassisShape = NULL;
	mCompoundShape = NULL;
	mVehicleRaycaster = NULL;
	mRaycastVehicle = NULL;

	mSteeringValue = 0.f;
	mMaxEngineForce = 1230; // This should be engine/velocity dependent
	mMaxBreakingForce = 25;
	mSteeringIncrement = 0.04f;
	mSteeringClamp = 0.4f;
	mEngineForce = 0.f;
	mBreakingForce = 0.f;

	mSuspensionStiffness = 180;//200.0f;
	mSuspensionDamping = 10;//50.0f;
	mSuspensionCompression = 10;//4.2f;
	mWheelFriction = 20;//1e30f;
	mRollInfluence = 0.1f;//1.0f;

#ifdef DEFAULT_VEHICLE_SETUP
	mWheelWidth = 0.1f;
	mWheelRadius = 0.13f;
#else
	mWheelWidth = 0.4f;
	mWheelRadius = 1.f;
#endif

	mSize = 1.0f;
}

Vehicle::~Vehicle(void)
{
	/*mDynamicsWorld->removeVehicle( mRaycastVehicle );
	SAFE_DELETE( mCompoundShape );
	SAFE_DELETE( mChassisShape );
	SAFE_DELETE( mVehicleRaycaster );
	SAFE_DELETE( mRaycastVehicle );
	SAFE_RELEASE( mWheelDebugMesh );
	SAFE_DELETE( mChassisMesh );
	SAFE_DELETE( mWheelMesh );
	SAFE_DELETE( mSimplifiedConvexShape );*/
}



bool Vehicle::Create(std::string chasis,std::string wheel,btDynamicsWorld *dynamicsWorld, btVector3 &position, btScalar mass )
{
	mDynamicsWorld = dynamicsWorld;

	//load car model
	{
		_carModel = new ModelObj2();
		_carModel->LoadModel(chasis);//"Assets/Models/Obj/willy/willy3.obj"
		_carModel->Optimize();
		//_carModel->Save("Assets/Models/Obj/willy/willy3.objb");
		_carModel->CreateVertexBuffers();
		_carModel->UseFrustumCheck(false);

	}

	//wheel model
	{
		_wheelModel = new ModelObj2();
		_wheelModel->LoadModel(wheel);
		_wheelModel->Optimize();
		//_wheelModel->Save("Assets/Models/Obj/willy/wheel3.objb");
		_wheelModel->CreateVertexBuffers();
		_wheelModel->UseFrustumCheck(false);

	}

#ifdef DEFAULT_VEHICLE_SETUP

	// create a chassis shape that is proportfion to the size	
	mChassisShape = new btBoxShape(btVector3(0.91f / 2.0f, 0.72f / 2.0f, 0.91f / 2.0f));
	mSize = 0.3f;
#else

	mSize = 1.7f;

	mChassisShape = new btConvexHullShape();

	// Extract vertices from vehicle mesh
	Vertex *v = 0;		
	mChassisMesh->GetMesh()->LockVertexBuffer(0, (void**)&v);
	for( DWORD i = 0; i < mChassisMesh->GetMesh()->GetNumVertices(); i++ )
	{
		((btConvexHullShape*)mChassisShape)->addPoint( btVector3( v[i].position.x, v[i].position.y, v[i].position.z ) );
	}
	mChassisMesh->GetMesh()->UnlockVertexBuffer();

#endif
	mCompoundShape = new btCompoundShape();
	//mWheelWidth = 0.4f * mSize;
	//mWheelRadius = 0.5f * mSize;

	btTransform localTrans;
	localTrans.setIdentity();
	localTrans.setOrigin( btVector3(0.0f,  (0.72f / 2.0f) + 0.135f, 0.0f ) );

#ifdef DEFAULT_VEHICLE_SETUP
	mCompoundShape->addChildShape(localTrans, mChassisShape);
#else
	mCompoundShape->addChildShape(localTrans, mSimplifiedConvexShape);
#endif

	localTrans.setOrigin( position );	
	mChassisRigidBody = CreateRigidBody( mass, localTrans, mCompoundShape );
	mChassisRigidBody->setCenterOfMassTransform( localTrans );
	mChassisRigidBody->setLinearVelocity(btVector3(0,0,0));
	mChassisRigidBody->setAngularVelocity(btVector3(0,0,0));	
	mDynamicsWorld->addRigidBody(mChassisRigidBody );

	mDynamicsWorld->getBroadphase()->getOverlappingPairCache()->cleanProxyFromPairs(mChassisRigidBody->getBroadphaseHandle(), mDynamicsWorld->getDispatcher());

	int rightIndex = 0;
	int upIndex = 1;
	int forwardIndex = 2;
	btVector3 wheelDirectionCS0(0,-1,0);
	btVector3 wheelAxleCS(-1,0,0);
	btScalar suspensionRestLength(0.5);

	mVehicleRaycaster = new btDefaultVehicleRaycaster(mDynamicsWorld);
	mRaycastVehicle = new btRaycastVehicle(mVehicleTuning, mChassisRigidBody, mVehicleRaycaster);				

	// Never deactivate the vehicle
	mChassisRigidBody->setActivationState(DISABLE_DEACTIVATION);

	mDynamicsWorld->addVehicle(mRaycastVehicle);

#ifdef DEFAULT_VEHICLE_SETUP
	float connectionHeight =  (0.72f / 2.0f) + 0.135f + 0.135f;	//0.535
#else
	float connectionHeight = 1.9f;	
#endif

	m_wheelShape = new btCylinderShapeX(btVector3(mWheelWidth,mWheelRadius,mWheelRadius));

	bool isFrontWheel = true;		
	//choose coordinate system
	mRaycastVehicle->setCoordinateSystem(rightIndex, upIndex, forwardIndex);

	btVector3 connectionPointCS0(mSize+(.5f*mWheelWidth),connectionHeight,1.4f*mSize-mWheelRadius);		
	mRaycastVehicle->addWheel(connectionPointCS0,wheelDirectionCS0,wheelAxleCS,suspensionRestLength,mWheelRadius,mVehicleTuning,isFrontWheel);

	connectionPointCS0 = btVector3(-mSize-(.5f*mWheelWidth),connectionHeight,1.4f*mSize-mWheelRadius);
	mRaycastVehicle->addWheel(connectionPointCS0,wheelDirectionCS0,wheelAxleCS,suspensionRestLength,mWheelRadius,mVehicleTuning,isFrontWheel);

	isFrontWheel = false;
	connectionPointCS0 = btVector3(-mSize-(.5f*mWheelWidth),connectionHeight,-1.6f*mSize+mWheelRadius);
	mRaycastVehicle->addWheel(connectionPointCS0,wheelDirectionCS0,wheelAxleCS,suspensionRestLength,mWheelRadius,mVehicleTuning,isFrontWheel);

	connectionPointCS0 = btVector3(mSize+(.5f*mWheelWidth),connectionHeight,-1.6f*mSize+mWheelRadius);
	mRaycastVehicle->addWheel(connectionPointCS0,wheelDirectionCS0,wheelAxleCS,suspensionRestLength,mWheelRadius,mVehicleTuning,isFrontWheel);

	for (int i=0;i<mRaycastVehicle->getNumWheels();i++)
	{
		btWheelInfo& wheel = mRaycastVehicle->getWheelInfo(i);
		wheel.m_suspensionStiffness = mSuspensionStiffness;
		wheel.m_wheelsDampingRelaxation = mSuspensionDamping;
		wheel.m_wheelsDampingCompression = mSuspensionCompression;
		wheel.m_frictionSlip = mWheelFriction;
		wheel.m_rollInfluence = mRollInfluence;			
	}

	return false;
}

void Vehicle::Render()
{
	Aurora::Graphics::RenderManager* _renderManager = Aurora::Graphics::RenderManager::Instance();

	btVector3 wheelColor(1,0,0);
	btVector3 carColor(0,0,1);

	btVector3	worldBoundsMin,worldBoundsMax;
	mDynamicsWorld->getBroadphase()->getBroadphaseAabb(worldBoundsMin,worldBoundsMax);

	// Draw wheels (cylinders)
	for (int i=0;i<mRaycastVehicle->getNumWheels();i++)
	{
		//synchronize the wheels with the (interpolated) chassis worldtransform
		mRaycastVehicle->updateWheelTransform(i, true );				

		float cubeMatrix[16];
		mRaycastVehicle->getWheelInfo(i).m_worldTransform.getOpenGLMatrix((btScalar*)&cubeMatrix);

		//m_shapeDrawer->drawOpenGL(cubeMatrix,m_wheelShape,wheelColor,0,worldBoundsMin,worldBoundsMax);

		_renderManager->PushMatrix();
		_renderManager->MultMatrix(cubeMatrix);
		
		_wheelModel->Draw();

		_renderManager->PopMatrix();

#ifndef DEFAULT_VEHICLE_SETUP
		mDevice->SetTransform( D3DTS_WORLD, &matWorld );
		if( mWheelMesh )
		{
			mWheelMesh->GetMesh()->DrawSubset( 0 );
		}
#endif
	
	}

	// Draw chassis
	{
		btTransform trn;
		mRaycastVehicle->getRigidBody()->getMotionState()->getWorldTransform( trn );		


	

#ifndef DEFAULT_VEHICLE_SETUP

		if( mChassisMesh )
		{
			D3DXMATRIX matWorld = ConvertMatrix( trn );		
			D3DXMATRIX matRot, matTrans, matScale;
			// offset 1.0 for the wheel
			D3DXMatrixTranslation( &matTrans, 0.0f, 1.0f, 0.0f );
			D3DXMatrixMultiply(&matWorld, &matTrans, &matWorld);		

			mDevice->SetTransform( D3DTS_WORLD, &matWorld );		
			mChassisMesh->GetMesh()->DrawSubset( 0 );
		}
#else
		float cubeMatrix[16];
		trn.getOpenGLMatrix((btScalar*)&cubeMatrix);

		//m_shapeDrawer->drawOpenGL(cubeMatrix,mChassisShape,carColor,0,worldBoundsMin,worldBoundsMax);

		_renderManager->PushMatrix();
		_renderManager->MultMatrix(cubeMatrix);
		
		_carModel->Draw();

		_renderManager->PopMatrix();
#endif			
	}
}

void Vehicle::Update(SystemManager* systemManager, float elapsedTime )
{
	if(systemManager->keyHold(Key::Right))
	{
		mSteeringValue -= mSteeringIncrement;
		if ( mSteeringValue < -mSteeringClamp)
			mSteeringValue = -mSteeringClamp;
	}else
	{
		
		// Return wheel to original position
		if ( mSteeringValue <= 0.0f )
		{
			mSteeringValue += mSteeringIncrement;

			if ( mSteeringValue > 0.0f )
				mSteeringValue = 0.0f;
		}

		
	}

	if(systemManager->keyHold(Key::Left))
	{
		mSteeringValue += mSteeringIncrement;
		if(mSteeringValue > mSteeringClamp)
			mSteeringValue = mSteeringClamp;
	}else
	{
		// Return wheel to original position
		if( mSteeringValue >= 0.0f )
		{
			mSteeringValue -= mSteeringIncrement;
			if ( mSteeringValue < 0.0f )
				mSteeringValue = 0.0f;
		}

		
	}

	if(systemManager->keyHold(Key::Up))
	{
		mEngineForce = mMaxEngineForce;
		mBreakingForce = 0.f;		
	}else
	{
		mEngineForce = 0.f;
	}

	if(systemManager->keyHold(Key::Down))
	{
		mEngineForce = -mMaxEngineForce;
		mBreakingForce = 0.f;
	}else
	{
		mBreakingForce = 0.f; 			
	}


	int wheelIndex = 2;
	mRaycastVehicle->applyEngineForce(mEngineForce,wheelIndex);
	mRaycastVehicle->setBrake(mBreakingForce,wheelIndex);
	wheelIndex = 3;
	mRaycastVehicle->applyEngineForce(mEngineForce,wheelIndex);
	mRaycastVehicle->setBrake(mBreakingForce,wheelIndex);

	wheelIndex = 0;
	mRaycastVehicle->setSteeringValue(mSteeringValue,wheelIndex);
	wheelIndex = 1;
	mRaycastVehicle->setSteeringValue(mSteeringValue,wheelIndex);		

	// Update vehicle direction
	btWheelInfo& wheel = mRaycastVehicle->getWheelInfo( 0 );
	btVector3 up = -wheel.m_raycastInfo.m_wheelDirectionWS;
	const btVector3& right = wheel.m_raycastInfo.m_wheelAxleWS;
	mForward = up.cross(right);
	mForward.normalize();
}

void Vehicle::Destroy()
{

}

btRigidBody* Vehicle::CreateRigidBody(float mass, const btTransform& startTransform, btCollisionShape* shape)
{
	//rigidbody is dynamic if and only if mass is non zero, otherwise static
	bool isDynamic = (mass != 0.f);

	btVector3 localInertia(0,0,0);
	if (isDynamic)
		shape->calculateLocalInertia(mass,localInertia);
	
	//using motionstate is recommended, it provides interpolation capabilities, and only synchronizes 'active' objects
	btDefaultMotionState* myMotionState = new btDefaultMotionState( startTransform );
	btRigidBody::btRigidBodyConstructionInfo cInfo(mass,myMotionState,shape,localInertia);
	btRigidBody* body = new btRigidBody(cInfo);

	return body;
}